# Copyright (C) 2012-2013 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import sys
import code
import readline
import __builtin__

from LMICompleter import LMICompleter
from LMIConnection import LMIConnection
from LMIShellConfig import LMIShellConfig
from LMIReturnValue import LMIReturnValue

from LMIConnection import connect

from LMIUtil import lmi_set_use_exceptions
from LMIUtil import lmi_get_use_exceptions

def _lmi_displayhook(o):
    """
    LMIShell display hook, which is used, when run in the interactive mode. Prints out a
    pretty message, when a LMIReturnValue is returned from a function, method call, etc.
    It is important, that the LMIReturnValue object needs to have a property
    "use_display_hook" set.
    """
    if o is None:
        return
    __builtin__._ = None
    if isinstance(o, LMIReturnValue) and o.use_display_hook:
        o.use_display_hook = False
        if (isinstance(o.rval, int) and o.rval in (0, 4096)) or (isinstance(o.rval, bool) and o.rval == True):
            result_str = "ok"
        else:
            result_str = "fail"
        sys.stdout.write("%s: %s (%s)" % (o.hostname, result_str, str(o.rval)))
        if o.errorstr:
            sys.stdout.write(": %s" % o.errorstr)
        sys.stdout.write("\n")
    else:
        print o
    __builtin__._ = o

sys.displayhook = _lmi_displayhook

def _use_display_sugar(use=True):
    """
    Helper function, which controls, if to use display "sugar". That means, if to use LMI
    display hook for LMIReturnValue.
    """
    sys.displayhook = _lmi_displayhook if use else sys.__displayhook__

class LMIConsole(code.InteractiveConsole):
    """
    Class representing an interactive console.
    """
    SHELL_PS1 = "> "
    SHELL_PS2 = "... "
    DEFAULT_LOCALS = {
        "LMIConnection" : LMIConnection,
        "use_exceptions" : lmi_set_use_exceptions
    }

    def __init__(self):
        """
        Constructs a LMIConsole object.
        """
        # Setup default LMIConsole locals, that are needed
        # for proper object construction
        locals = dict(LMIConsole.DEFAULT_LOCALS.items())
        code.InteractiveConsole.__init__(self, locals)

        # Store configuration
        config = LMIShellConfig()
        self._history_file = config.history_file
        self._history_length = config.history_length
        self._use_cache = config.use_cache
        self._use_exceptions = config.use_exceptions

        # Setup LMIShell-wide option, which defines, whether the LMIShell
        # should propagate caught exceptions, or dump them
        lmi_set_use_exceptions(self._use_exceptions)

    def setup_completer(self):
        """
        Initializes tab-completer.
        """
        self._completer = LMICompleter(self.locals)
        readline.set_completer(self._completer.complete)
        readline.parse_and_bind("tab: complete")

    def interact(self, locals=None):
        """
        Starts the interactive mode.

        Arguments:
            locals -- dictionary with locals
        """
        # Initialize the interpreter
        if locals is None:
            locals = {}
        else:
            for (k, v) in locals.iteritems():
                if isinstance(v, LMIConnection):
                    locals[k]._client.interactive = True
        locals["clear_history"] = self.clear_history
        locals["use_display_sugar"] = _use_display_sugar
        locals["connect"] = lambda h, u="", p="": connect(h, u, p, True, self._use_cache)
        self.locals = dict(self.locals.items() + locals.items())
        self.setup_completer()
        self.load_history()
        old_sys_ps1 = ""
        try:
            old_sys_ps1 = sys.ps1
        except AttributeError, e:
            old_sys_ps1 = ""
        try:
            old_sys_ps2 = sys.ps2
        except AttributeError, e:
            old_sys_ps2 = ""
        sys.ps1 = LMIConsole.SHELL_PS1
        sys.ps2 = LMIConsole.SHELL_PS2

        # Interact
        more = 0
        while 1:
            try:
                if more:
                    prompt = sys.ps2
                else:
                    prompt = sys.ps1
                try:
                    line = self.raw_input(prompt)
                    # Can be None if sys.stdin was redefined
                    encoding = getattr(sys.stdin, "encoding", None)
                    if encoding and not isinstance(line, unicode):
                        line = line.decode(encoding)
                except EOFError:
                    self.write("\n")
                    break
                else:
                    more = self.push(line)
            except KeyboardInterrupt:
                self.write("\n")
                self.resetbuffer()
                more = 0

        # Cleanup after the interpreter
        if old_sys_ps1:
            sys.ps1 = old_sys_ps1
        if old_sys_ps2:
            sys.ps2 = old_sys_ps2
        self.save_history()

    def interpret(self, script_name, script_argv, locals=None):
        """
        Interprets a specified script within additional provided locals.
        There are LMIConsole.DEFAULT_LOCALS present.

        Arguments:
            script_name -- string containing a script name
            script_argv -- script CLI arguments
            locals -- dictionary with locals
        """
        # Initialize locals
        if locals is None:
            locals = {}
        else:
            for (k, v) in locals.iteritems():
                if isinstance(v, LMIConnection):
                    locals[k]._client.interactive = False
        locals["connect"] = lambda h, u="", p="": connect(h, u, p, False, self._use_cache)
        self.locals = dict(self.locals.items() + locals.items())

        # Execute the script
        sys.argv = script_argv
        exit_code = 0
        try:
            execfile(script_name, self.locals)
        except SystemExit, e:
            exit_code = e.code

        # Return with exit_code
        return exit_code

    def load_history(self):
        """
        Loads the shell's history from the history file.
        """
        if self._history_length == 0 or not os.path.exists(self._history_file):
            return
        readline.read_history_file(self._history_file)
        if self._history_length > 0 and readline.get_current_history_length() > self._history_length:
            readline.set_history_length(self._history_length)
            readline.write_history_file(self._history_file)
            readline.read_history_file(self._history_file)

    def save_history(self):
        """
        Saves current history of commands into the history file. If the length of history
        exceeds a maximum history file length, the history will be truncated.
        """
        if self._history_length == 0:
            return
        elif self._history_length > 0:
            readline.set_history_length(self._history_length)
        try:
            readline.write_history_file(self._history_file)
        except IOError, e:
            pass

    def clear_history(self):
        """
        Clears the current history.
        """
        readline.clear_history()
