# Copyright (C) 2012-2013 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import collections

from LMIExceptions import LMIReturnValueError

_RValue = collections.namedtuple('RValue', ["rval", "rparams", "errorstr"])

class LMIReturnValue(_RValue):
    """
    Class representing a return value, which holds 3 main types of attributes:
        rval -- return value
        rparams -- returned parameters of eg. method call
        errorstr -- error string
    """
    def __getattr__(self, name):
        """
        Returns a property value, if found.

        Arguments:
            name -- string of property name
        """
        if not name in self._properties:
            raise AttributeError(name)
        return self._properties[name]

    def properties(self):
        """
        Returns a list of strings of all available properties.
        """
        return self._properties.keys()

    @property
    def use_display_hook(self):
        """
        Property returning a bool flag, if the LMIReturnValue can be used in the
        interactive mode's output hook.
        """
        return self._properties["use_display_hook"]

    @use_display_hook.setter
    def use_display_hook(self, value):
        """
        Property setter for display hook.
        """
        self._properties["use_display_hook"] = value

    @property
    def hostname(self):
        """
        Property returning a string of the hostname. Used in the interactive mode's output
        hook.
        """
        return self._properties["hostname"]

    def __new__(self, **kwargs):
        """
        Creates a new instance of LMIReturnValue. At least rval keyword argument must be
        provided to properly construct a LMIReturnValue object; otherwise
        LMIReturnValueError exception will be raised.

        Keyword arguments:
            rval -- return value
            rparams -- return parameters
            errorstr -- error string
            hostname -- ; default value is "This machine"
            use_display_hook -- ; default value is False

        Any other keyword argument will be also stored and is accessible via getattr.
        """
        if not "rval" in kwargs:
            raise LMIReturnValueError("rval argument missing while instancing LMIReturnValue object")
        properties = {
            "hostname" : "This Machine",
            "use_display_hook" : False
        }
        keys_to_pop = []
        for (key, value) in kwargs.iteritems():
            if key in ("rval", "rparams", "errorstr"):
                continue
            properties[key] = value
            keys_to_pop.append(key)
        for key in keys_to_pop:
            kwargs.pop(key)
        rval = _RValue.__new__(self,
            rval = kwargs["rval"],
            rparams = kwargs["rparams"] if "rparams" in kwargs else {},
            errorstr = kwargs["errorstr"] if "errorstr" in kwargs else "")
        rval._properties = properties
        return rval
