# Copyright (C) 2012-2013 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import sys
import pywbem

from LMIObjectFactory import LMIObjectFactory

# By default, the LMIShell does not use exceptions; LMIReturnValue
# is used instead (with proper error string).
class LMIUseExceptionsHelper(object):
    """
    Singleton helper class used for storing a bool flag, which defines,
    if the LMIShell should propagate exceptions or dump them.
    """
    _instance = None

    def __new__(cls):
        """
        Return a new LMIUseExceptionsHelper instance, if no shared object
        is present; otherwise an existing instance is returned. By default
        the "use_exceptions" flag is set to False.
        """
        if cls._instance is None:
            cls._instance = super(LMIUseExceptionsHelper, cls).__new__(cls)
            cls._instance._use_exceptions = False
        return cls._instance

    @property
    def use_exceptions(self):
        """
        Property returning a bool flag, which indicates, if the LMIShell should
        propagate the exceptions, or dump them.
        """
        return self._use_exceptions

    @use_exceptions.setter
    def use_exceptions(self, use=True):
        """
        Property setter, which modifies the bool flag, which indicates, if the LMIShell
        should propagate the exceptions, or dump them.
        """
        self._use_exceptions = use

class LMIPassByRef(object):
    """
    Helper class used for passing a value by reference. It uses the advantage of python,
    where all the dictionaries are passed by reference.

    Example of usage:
        by_ref = LMIPassByRef(some_value)
        by_ref.value == some_value
    """
    def __init__(self, val):
        """
        Constructs a LMIPassByRef object.

        Arguments:
            val -- value, which will be passed by reference
        """
        self._val = {0 : val}

    @property
    def value(self):
        """
        Property returning the value passed by reference.
        """
        return self._val[0]

    @value.setter
    def value(self, new_val):
        """
        Property setter for the value passed by reference.
        """
        self._val[0] = new_val

def lmi_get_item_or_default(dic, item, default):
    """
    Helper function, which returns either the dictionary member, or specified default
    value.

    Arguments:
        dic -- dictionary
        item -- dictionary key
        default -- default return value, if item is not present in the dictionary
    """
    return dic[item] if item in dic else default

def lmi_get_use_exceptions():
    """
    Returns a global flag indicating, if the LMIShell should use the exceptions,
    or dump them.
    """
    return LMIUseExceptionsHelper().use_exceptions

def lmi_set_use_exceptions(use=True):
    """
    Sets a global flag indicating, if the LMIShell should use the exceptions,
    or dump them.
    """
    LMIUseExceptionsHelper().use_exceptions = use

def lmi_raise_or_dump_exception(e=None):
    """
    Function which either raises an exception, or throws it away.
    """
    if not lmi_get_use_exceptions():
        return
    (et, ei, tb) = sys.exc_info()
    if e is None:
        raise et, ei, tb
    else:
        raise type(e), e, tb

def lmi_cast_to_cim(t, value):
    """
    Casts the value to CIM type.

    Arguments:
        t -- string of CIM type
        value -- variable to cast
    """
    cast = {
        "sint8"  : lambda x: pywbem.Sint8(x),
        "uint8"  : lambda x: pywbem.Uint8(x),
        "sint16" : lambda x: pywbem.Sint16(x),
        "uint16" : lambda x: pywbem.Uint16(x),
        "sint32" : lambda x: pywbem.Sint32(x),
        "uint32" : lambda x: pywbem.Uint32(x),
        "sint64" : lambda x: pywbem.Sint64(x),
        "uint64" : lambda x: pywbem.Uint64(x),
        "reference": lambda x: x.path if isinstance(x, LMIObjectFactory().LMIInstance) else x
    }
    cast_func = cast[t] if t.lower() in cast else lambda x: x
    if isinstance(value, list):
        return map(cast_func, value)
    elif isinstance(value, tuple):
        return tuple(map(cast_func, value))
    return cast_func(value)

def lmi_wrap_cim_namespace(client, cim_namespace_name):
    """
    Helper function, which returns wrapped CIMNamespace into LMINamespace.

    Arguments:
        client -- LMIShellClient object used for CIMOM communication
        cim_namespace_name -- string containing CIMNamespace name
    """
    return LMIObjectFactory().LMINamespace(client, cim_namespace_name)

def lmi_wrap_cim_class(client, cim_class_name, cim_namespace_name):
    """
    Helper function, which returns wrapped CIMClass into LMIClass.

    Arguments:
        client -- LMIShellClient object used for CIMOM communication
        cim_class_name -- string containing CIMClass name
        cim_namespace_name -- string containing CIMNamespace name,
             or None, if the namespace is not known
    """
    lmi_namespace = None
    if cim_namespace_name:
        lmi_namespace = lmi_wrap_cim_namespace(client, cim_namespace_name)
    return LMIObjectFactory().LMIClass(client, lmi_namespace, cim_class_name)

def lmi_wrap_cim_instance(client, cim_instance, cim_class_name, cim_namespace_name):
    """
    Helper function, which returns wrapped CIMInstance into LMIInstance.

    Arguments:
        client -- LMIShellClient object used for CIMOM communication
        cim_instance -- CIMInstance object to be wrapped
        cim_class_name -- string containing CIMClass name
        cim_namespace_name -- string containing CIMNamespace name,
            or None, if the namespace is not known
    """
    lmi_class = lmi_wrap_cim_class(client, cim_class_name, cim_namespace_name)
    return LMIObjectFactory().LMIInstance(client, lmi_class, cim_instance)

def lmi_wrap_cim_instance_name(client, cim_instance_name):
    """
    Helper function, which returns wrapped CIMInstanceName into LMIInstanceName.

    Arguments:
        client -- LMIShellClient object used for CIMOM communication
        cim_instance_name -- CIMInstanceName object to be wrapped
    """
    return LMIObjectFactory().LMIInstanceName(client, cim_instance_name)

def lmi_wrap_cim_method(client, cim_method_name, lmi_instance, sync_method):
    """
    Helper function, which returns wrapped CIMMethod imto LMIMethod.

    Arguments:
        client -- LMIShellClient object used for CIMOM communication
        cim_method_name -- string containing a method name
        lmi_instance -- LMIInstance object, on which the method call will be issued
        sync_method -- bool flag indicating, if we are trying to perform
            a synchronous method call
    """
    return LMIObjectFactory().LMIMethod(client, lmi_instance, cim_method_name, sync_method)

def lmi_transform_to_lmi(client, value):
    """
    Transforms returned values from a method call into LMI wrapped objects. Returns
    transformed input, where CIMInstance and CIMInstanceName are wrapped into LMI
    wrapper classes and primitive types are cast to python native types.

    Arguments:
        client -- LMIShellClient object used for CIMOM communication
        value -- object to be transformed into python type from pywbem one
    """
    if isinstance(value, pywbem.cim_obj.CIMInstance):
        namespace = value.path.namespace if value.path else None
        return lmi_wrap_cim_instance(client, value, value.classname, namespace)
    elif isinstance(value, pywbem.cim_obj.CIMInstanceName):
        return lmi_wrap_cim_instance_name(client, value)
    elif isinstance(value, pywbem.CIMInt):
        return int(value)
    elif isinstance(value, pywbem.CIMFloat):
        return float(value)
    elif isinstance(value, dict):
        return dict(
            zip(value.keys(),
            map(lambda val: lmi_transform_to_lmi(client, val), value.values()))
        )
    elif isinstance(value, pywbem.NocaseDict):
        return pywbem.NocaseDict(
            zip(value.keys(),
            map(lambda val: lmi_transform_to_lmi(client, val), value.values()))
        )
    elif isinstance(value, list):
        return map(lambda val: lmi_transform_to_lmi(client, val), value)
    elif isinstance(value, tuple):
        return tuple(map(lambda val: lmi_transform_to_lmi(client, val), value))
    return value
